@@ -1,7 +1,7 @@
 /*
     Numdiff - compare putatively similar files, 
     ignoring small numeric differences
-    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010  Ivano Primi  <ivprimi@libero.it>
+    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  Ivano Primi  <ivprimi@libero.it>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -25,18 +25,126 @@
 #include<locale.h>
 #endif
 #include"getopt.h"
-#include"numdiff.h" /* Only for LINE_INTERR definition */
+#include"numdiff.h" /* For IFS, NEWLINE and LINE_INTERR definitions */
 #include"ndselect.h"
 
 #include"read_line.c"
 
-static int scan_file (const Argslist* data)
+/*
+  This function has been copied from options.c,
+  since I do NOT want to add more dependencies to
+  the building of ndselect.
+*/
+static
+char* return_ifs (const char* optarg)
+{
+  char *s = (char*) malloc ((strlen(optarg) + 1) * sizeof(char));
+  
+  if(!s)
+    return NULL;
+  else
+    {
+      char *t, *u;
+
+      strcpy (s, optarg);
+      for (t = s; *t != '\0'; t++)
+	{
+	  if (*t == '\\')
+	    {
+	      switch (*(t+1))
+		{
+		case 'f':
+		  *t = '\f';
+		  break;
+		case 'n':
+		  *t = '\n';
+		  break;
+		case 'r':
+		  *t = '\r';
+		  break;
+		case 't':
+		  *t = '\t';
+		  break;
+		case 'v':
+		  *t = '\v';
+		  break;
+		default:
+		  *t = *(t+1);
+		}
+	      for (u = t+1; *u != '\0'; *u = *(u+1), u++);
+	    }
+	}
+      return s;
+    }
+}
+
+static 
+int we_can_go_on (const char *field, 
+		  unsigned long fieldno, unsigned long last_field)
+{
+  if (!last_field)
+    return (*field != '\0');
+  else
+    return (fieldno <= last_field && *field != '\0');
+}
+
+static
+void print_line (const char *line, const char *ifs,
+		 unsigned long first_field, unsigned long last_field,
+		 unsigned long increment, int omit_if_empty)
+{
+  unsigned long fieldno, lf;
+  int print_nl, at_least_one_field_printed;
+  const char *field, *endfield, *nextfield, *stop, *ptr;
+  
+  if (!last_field)
+    lf = 0;
+  else
+    {
+      lf = (last_field - first_field) / increment;
+      lf = lf * increment + first_field;
+    }
+  at_least_one_field_printed = 0;
+  print_nl = strchr (line, NEWLINE) != NULL ? 1 : 0;
+  field = line + strspn (line, ifs);
+  for (fieldno = 1; (we_can_go_on (field, fieldno, last_field)) ; fieldno++)
+    {
+      endfield = field + strcspn (field, ifs);
+      nextfield = endfield + strspn (endfield, ifs);
+      if (fieldno >= first_field && (fieldno - first_field) % increment == 0)
+	{
+	  stop = (fieldno == lf || *nextfield == '\0') ? endfield : nextfield;
+	  for (ptr = field; ptr < stop; ptr++)
+	    fputc(*ptr, stdout);
+	  at_least_one_field_printed = 1;
+	}
+      field = nextfield;
+    }
+  if ((omit_if_empty))
+    {
+      if ((print_nl) && (at_least_one_field_printed))
+	fputc(NEWLINE, stdout);
+    }
+  else
+    {
+      if ((print_nl))
+	fputc(NEWLINE, stdout);
+      else
+	fputc(' ', stdout);
+    }
+}
+
+static
+int scan_file (const Argslist* data)
 {
   FILE *fp;
   char *line_buffer;
+  char *ifs;
   unsigned long lineno;
-  int errcode;
+  int errcode, omit_empty_lines;
 
+  ifs = (!data->ifs) ? IFS : data->ifs;
+  omit_empty_lines = (data->optmask & ___X_MASK) ? 1 : 0;
   if (!data->file || !*data->file)
     fp = stdin;
   else
@@ -52,7 +160,7 @@ static int scan_file (const Argslist* data)
 	   lineno++)
 	{
 	  if (lineno >= data->begin_line && (lineno - data->begin_line) % data->step == 0)
-	    fputs (line_buffer, stdout);
+	    print_line (line_buffer, ifs, data->first_field, data->last_field, data->increment, omit_empty_lines);
 	  free ((void*)line_buffer);
 	}
     }
@@ -64,7 +172,7 @@ static int scan_file (const Argslist* data)
 	   lineno++)
 	{
 	  if (lineno >= data->begin_line && (lineno - data->begin_line) % data->step == 0)
-	    fputs (line_buffer, stdout);
+	    print_line (line_buffer, ifs, data->first_field, data->last_field, data->increment, omit_empty_lines);
 	  free ((void*)line_buffer);
 	}
     }
@@ -86,10 +194,11 @@ static int scan_file (const Argslist* data)
     }
 }
 
-static void print_selversion (const char* progname)
+static
+void print_selversion (const char* progname)
 {
   printf ("%s %s\n", progname, VERSION);
-  printf ("Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010  %s <ivprimi@libero.it>\n", 
+  printf ("Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  %s <ivprimi@libero.it>\n", 
 	  /* TRANSLATORS: This is a proper name.  See the gettext
 	     manual, section Names.
 	     Pronounciation is like "evaa-no pree-me".  */
@@ -101,57 +210,72 @@ This is free software: you are free to change and redistribute it.\n\
 There is NO WARRANTY, to the extent permitted by law.\n"));
 }
 
-static void print_selhelp (const char* progname)
+static
+void print_selhelp (const char* progname)
 {
   puts (_("Usage:"));
   printf ("%s -h|--help|-v|--version   %s\n\n", progname, _("or"));
-  printf ("%s %s\n", progname, "[-b FIRST][-e LAST][-s STEP][-l PATH][-o PATH] [FILE]");
+  printf ("%s %s\n", progname, "[-b N][-e N][-s N][-F N][-L N][-I N][-S IFS][-x][-l PATH][-o PATH] [FILE]");
   /* %%% */
-  printf(_("\n\
-For a given range of line numbers and a given step\n\
-print on the standard output all lines of a file,\n\
-starting with the first line of the range and ending within\n\
-its last line, whose line number is such that the difference\n\
-between it and the start point is an integer multiple\n\
-of the given step\n\n"));
-  printf ("-b, --beginning, --start=FIRST\n    %s\n    %s\n", 
-	  _("Specify the first line to print"),
+  printf(_("\nPrint to standard output a subset of lines and fields from a given file.\n"));
+  printf ("\n%s\n%s\n%s\n\n%s\n\n",
+	  _("The argument after the options is the name of the file to read from."),
+	  _("The complete path of the file should be given,\na directory name is not accepted."),
+	  _("If no input file is specified, the program reads from the standard input."),
+	  _("Exit status: 0 in case of normal termination, -1 (255) in case of error"));
+  printf ("-b, --beginning, --start=N\n    %s\n    %s\n", 
+	  _("Set to N the number of the first line to print"),
 	  _("(The default behavior is to start with line number 1)"));
-  printf ("-e, --end=LAST\n    %s\n    %s\n", 
-	  _("Specify the last line that can be printed"),
+  printf ("-e, --end=N\n    %s\n    %s\n", 
+	  _("Set to N the number of the last line that can be printed"),
 	  _("(The default behavior is to arrive till to the end of the file)"));
-  printf ("-s, --step=STEP\n    %s\n    %s\n", 
-	  _("Specify the step to use when selecting the lines to print"),
-	  _("(The default value for the step is 1)"));
+  printf ("-s, --step=N\n    %s\n    %s\n", 
+	  _("Set to N the increment to use when selecting the lines to print"),
+	  _("(The default value for the increment is 1)"));
+  printf ("-F, --first-field=N\n    %s\n    %s\n", 
+	  _("Set to N the number of the first field to print"),
+	  _("(The default behavior is to start with field number 1)"));
+  printf ("-L, --last-field=N\n    %s\n    %s\n", 
+	  _("Set to N the number of the last field that can be printed"),
+	  _("(The default behavior is to arrive till to the end of every line)"));
+  printf ("-I, --increment=N\n    %s\n    %s\n", 
+	  _("Set to N the increment to use when selecting the fields to print"),
+	  _("(The default value for the increment is 1)"));
+  printf ("-S, --separator=IFS\n    %s\n    %s\n",
+          _("Specify the set of characters to use\n    to split the input lines into fields"),
+          _("(The default set of characters is space, tab and newline)."));
+  printf ("-x, --omit-empty-lines\n    %s\n",
+	  _("Do not print empty lines"));
   printf ("-l, --warnings-to=PATH\n    %s\n",
 	  _("Redirect warning and error messages from stderr to the indicated file"));
   printf ("-o, --output=PATH\n    %s\n",
 	  _("Redirect output from stdout to the indicated file"));
   printf ("-h, --help\n    %s\n", _("Show this help message"));
-  printf ("-v, --version\n    %s\n", _("Show version number, Copyright, Distribution Terms and NO-Warranty"));
-  printf ("\n%s\n%s\n%s\n%s\n\n",
-	  _("The argument after the options is the name of the file to scan."),
-	  _("The complete path of the file should be given,\na directory name is not accepted."),
-	  _("If no input file is specified, the program reads from the standard input."),
-	  _("Exit status: 0 in case of normal termination, -1 (255) in case of error"));
+  printf ("-v, --version\n    %s\n\n", _("Show version number, Copyright, Distribution Terms and NO-Warranty"));
 }
 
 extern int errno;
 extern char *optarg; 
 extern int optind;
 
-static int set_args (int argc, char* argv[], Argslist *list)
+static
+int set_args (int argc, char* argv[], Argslist *list)
 {
-  const char *optstring = "hb:e:s:l:o:v";
+  const char *optstring = "hb:e:s:F:L:I:S:xl:o:v";
   struct option long_options[] = {
-    {"help",         0, NULL, 'h'},
-    {"beginning",    1, NULL, 'b'},
-    {"start",        1, NULL, 'b'},
-    {"end",          1, NULL, 'e'},
-    {"step",         1, NULL, 's'},
-    {"warnings-to",  1, NULL, 'l'},
-    {"output",       1, NULL, 'o'},
-    {"version",      0, NULL, 'v'},
+    {"help",             0, NULL, 'h'},
+    {"beginning",        1, NULL, 'b'},
+    {"start",            1, NULL, 'b'},
+    {"end",              1, NULL, 'e'},
+    {"step",             1, NULL, 's'},
+    {"first-field",      1, NULL, 'F'},
+    {"last-field",       1, NULL, 'L'},
+    {"increment",        1, NULL, 'I'},
+    {"separator",        1, NULL, 'S'},
+    {"omit-empty-lines", 0, NULL, 'x'},
+    {"warnings-to",      1, NULL, 'l'},
+    {"output",           1, NULL, 'o'},
+    {"version",          0, NULL, 'v'},
     {0, 0, 0, 0}
   };
   int option_index=0;
@@ -166,6 +290,10 @@ static int set_args (int argc, char* argv[], Argslist *list)
   list->begin_line=1;
   list->end_line=0;
   list->step=1;
+  list->first_field=1;
+  list->last_field=0;
+  list->increment=1;
+  list->ifs = NULL;
   list->file=NULL;
 
   while ( (optch = getopt_long (argc, argv, optstring, long_options, &option_index)) != -1 )
@@ -217,6 +345,68 @@ static int set_args (int argc, char* argv[], Argslist *list)
 	      return -1;
 	    }
 	  break;
+	case 'F':
+	  list->optmask |= ___SF_MASK;
+	  errno = 0;
+	  for (endptr = optarg; is_space (*endptr) != 0; endptr++);
+	  if (*endptr == '-' ||
+	      (list->first_field = strtoul (optarg, &endptr, 10), 
+	       errno == ERANGE) || list->first_field == 0 ||
+	      *endptr != '\0')
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+		       PACKAGE2, optch);
+	      return -1;
+	    }
+	  break;
+	case 'L':
+	  list->optmask |= ___SL_MASK;
+	  errno = 0;
+	  for (endptr = optarg; is_space (*endptr) != 0; endptr++);
+	  if (*endptr == '-' ||
+	      (list->last_field = strtoul (optarg, &endptr, 10), 
+	       errno == ERANGE) ||
+	      *endptr != '\0')
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+		       PACKAGE2, optch);
+	      return -1;
+	    }
+	  break;
+	case 'I':
+	  list->optmask |= ___SI_MASK;
+	  errno = 0;
+	  for (endptr = optarg; is_space (*endptr) != 0; endptr++);
+	  if (*endptr == '-' ||
+	      (list->increment = strtoul (optarg, &endptr, 10), 
+	       errno == ERANGE) || list->increment == 0 ||
+	      *endptr != '\0')
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+		       PACKAGE2, optch);
+	      return -1;
+	    }
+	  break;
+        case 'S':
+          if ((list->ifs))
+	    {
+	      free((void*)list->ifs);
+	      list->ifs = NULL;
+	    }
+          list->ifs = return_ifs (optarg);
+	  if ( ((list->ifs) && !strchr(list->ifs, NEWLINE)) )
+            {
+              fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
+                       PACKAGE, optch);
+              fprintf (stderr, _("  A list of field delimiters can not be empty and\n  must always include the newline character (\'\\n\')\n"));
+              return -1;
+            }
+          else
+            list->optmask |= ___SS_MASK;
+          break;
+	case 'x':
+	  list->optmask |= ___X_MASK;
+	  break;
 	case 'l':
 	  if (!freopen (optarg, "w", stderr))
 	    {
