@@ -1,7 +1,7 @@
 /*
     Numdiff - compare putatively similar files, 
     ignoring small numeric differences
-    Copyright (C) 2005, 2006, 2007, 2008, 2009  Ivano Primi  <ivprimi@libero.it>
+    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  Ivano Primi  <ivprimi@libero.it>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,6 +22,10 @@
 
 #include "system.h"
 
+#if defined(HAVE_LIBGMP) && !defined(DISABLE_GMP)
+#define USE_GMP 1
+#endif /* defined(HAVE_LIBGMP) && !defined(DISABLE_GMP) */
+
 /* The type of a hash value.  */
 typedef size_t hash_value;
 verify (hash_value_is_unsigned, ! TYPE_SIGNED (hash_value));
@@ -30,7 +34,7 @@ verify (hash_value_is_unsigned, ! TYPE_SIGNED (hash_value));
 #define ROL(v, n) ((v) << (n) | (v) >> (sizeof (v) * CHAR_BIT - (n)))
 
 /* Given a hash value and a new character, return a new hash value.  */
-#ifdef _DEBUG_
+#ifdef _DEBUG_HASHING_
 #define HASH(h, c) (putc((c), stderr), (c) + ROL (h, 7))
 #else
 #define HASH(h, c) ((c) + ROL (h, 7))
@@ -52,16 +56,68 @@ verify (hash_value_is_unsigned, ! TYPE_SIGNED (hash_value));
 
 #include"number.h"
 
+#ifdef USE_GMP
+
+#include<gmp.h>
+
+typedef mpf_t Real;
+typedef struct {
+  mpf_t re, im;
+} Complex;
+
+#else /* not USE_GMP */
+
 typedef bc_num Real;
 typedef struct {
   bc_num re, im;
 } Complex;
 
+#endif /* USE_GMP */
+
 /*
   End section -- Math types
 */
 
+#define THRLIST_OK              0
+#define THRLIST_INVALID_FORMAT -1
+#define THRLIST_INVALID_RANGES -2
+
+/*
+  A structure of type 'thrlist_node' is used
+  to construct lists of threshold values.
+*/
+
+struct __thrlist_node {
+  Real threshold;               /* the threshold value */
+
+  /* BEG1-END1 and BEG2-END2 are the ranges of fields    */
+  /* to which the specified THRESHOLD value applies.     */
+  /* The two ranges must have the same length, but refer */
+  /* to different files.                                 */
+
+  unsigned long beg1, end1;     
+  unsigned long beg2, end2;
+
+  /* DOUBLE_RANGE_SPEC is a boolean value, which is 1 if */
+  /* BEG1-END1 and BEG2-END2  have been both explicitely */
+  /* set via the command line through a specification of */
+  /* the form:                                           */
+  /*                 BEG1-END1:BEG2-END2                 */
+  /* and otherwise is equal to 0.                        */
+
+  int double_range_spec;
+
+  /* pointer to the next node in the list */
+  struct __thrlist_node *next;  
+};
+
+typedef struct __thrlist_node thrlist_node;
+
+/* A pointer to a list of threshold values */
+typedef thrlist_node *thrlist;
+
 struct numfmt {
+  char *currency;    /* currency            */
   char dp;           /* decimal point       */
   char thsep;        /* thousands separator */
   unsigned grouping; /* Number of digits in each group */
@@ -76,10 +132,11 @@ struct numfmt {
 
 /*
   This is the number of the fields in the 'numftm' structure
-  having "char" type. They must have all different values
-  (see the code of the function valid_numfmt() in the file options.c) .
+  having "char" type. None of them can be a digit and 
+  they must have all different values
+  (see the code of the function valid_numfmt() in the file options.c).
 */
-#define NUMFMT_CHARS 6 
+#define NUMFMT_CHARS 6
 
 typedef struct {
   unsigned char* ptr;
@@ -131,8 +188,8 @@ typedef struct {
   /* WR_TO_SECOND_FILE          2                     */
   int relerr_formula;
 
-  /* Tolerance for absolute and relative errors */
-  Real maxabserr, maxrelerr;
+  /* Tolerance thresholds for absolute and relative errors */
+  thrlist maxabserr, maxrelerr;
 
   /* These variables are used to print statistics */
   Real Labserr,  Crelerr,  Lrelerr,  Cabserr, N1abserr, N1disperr, N2abserr, N2disperr;
@@ -169,7 +226,7 @@ typedef struct {
 #define _H_MASK   0x00000001 /* -h option, used to recall help */
 #define _A_MASK   0x00000002 /* -a option, used to set tolerance for abs. error  */
 #define _R_MASK   0x00000004 /* -r option, used to set tolerance for rel. error  */
-#define _2_MASK   0x00000008 /* -2 option, used to enable the "flexible" control */
+#define _2_MASK   0x00000008 /* -2 option, used to enable the "strict" control */
 #define _S_MASK   0x00000010 /* -s option, used to explicitly set IFS  */
 #define _B_MASK   0x00000020 /* -b option, used to enable the "brief" mode */
 #define _F_MASK   0x00000040 /* -f option, used to enable the "filter-only" mode */
@@ -202,29 +259,31 @@ typedef struct {
 #define _SE_MASK  0x01000000 /* -E option, used to enable the "essential" 
 				mode */
 #define _SV_MASK  0x02000000 /* -V option, used to enable the "verbose" mode */
-#define _SS_MASK  0x04000000 /* -S option, used to print statistics */
-#define _SI_MASK  0x08000000 /* -I option, used to ignore case while comparing
+#define _SO_MASK  0x04000000 /* -O option, used to enable the "overview" mode */
+#define _SS_MASK  0x08000000 /* -S option, used to print statistics */
+#define _SI_MASK  0x10000000 /* -I option, used to ignore case while comparing
 			        non numerical fields */
-#define _SH_MASK  0x10000000 /* -H option, by filtering assume large files and
+#define _SH_MASK  0x20000000 /* -H option, by filtering assume large files and
 				many scattered small changes */
-#define _M_MASK   0x20000000 /* -m option, by filtering try hard to find a smaller 
+#define _M_MASK   0x40000000 /* -m option, by filtering try hard to find a smaller 
 				set of changes */
-#define _V_MASK   0x40000000 /* -v option, used to show version number,
+#define _V_MASK   0x80000000 /* -v option, used to show version number,
 			        Copyright and No-Warrany */
+/* Remark: Intentionally there are no masks for the options -c and -F */
 
 /* Output modes: verbose, normal, brief, and quiet.       */
 /* Do not change the relative order of the values of      */
 /* these macros, the code in cmp_lines()  (see file       */
 /* cmpfns.c) relies on the fact that:                     */
 /* OUTMODE_VERBOSE > OUTMODE_NORMAL > OUTMODE_COINCISE    */
-/* > OUTMODE_BRIEF > OUTMODE_QUIET  .                     */
+/* > OUTMODE_BRIEF > OUTMODE_QUIET  > OUTMODE_OVERVIEW    */
 
 #define OUTMODE_VERBOSE     4
 #define OUTMODE_NORMAL      3
 #define OUTMODE_COINCISE    2
 #define OUTMODE_BRIEF       1
 #define OUTMODE_QUIET       0
-
+#define OUTMODE_OVERVIEW   -1
 
 /* Methods to compute the relative differences */
 
@@ -244,12 +303,21 @@ typedef struct {
 
 #define CHAR_ZERO '0'
 
+/* The character representing the number one */
+
+#define CHAR_ONE '1'
+
+/* The character representing the number nine */
+
+#define CHAR_NINE '9'
+
 /* newline character */
 
 #define NEWLINE '\n'
 
 /*
   Predefined values for
+  .nf*.currency (currency name)
   .nf*.dp       (decimal point)
   .nf*.thsep    (thousands separator)
   .nf*.grouping (number of digits in each thousands group)
@@ -260,6 +328,7 @@ typedef struct {
   .iscale      (decimal digits of accuracy)
   .ifs         (internal fields separator)
 */
+#define CURRENCY  ""
 #define DP        '.'
 #define THSEP     ','
 #define GROUPING   3
@@ -280,13 +349,13 @@ typedef struct {
   Largest possible exponent accepted by Numdiff
   when a number is written in scientific notation
 */
-#define MAX_EXPN    +1073741824
+#define MAX_EXPN    +1073741824L
 
 /*
   Lowest possible exponent accepted by Numdiff
   when a number is written in scientific notation
 */
-#define MIN_EXPN    -1073741824
+#define MIN_EXPN    -1073741824L
 
 /*
   Macro to move ahead a pointer
@@ -294,34 +363,36 @@ typedef struct {
 #define move_ahead(ptr) ptr++
 
 /*
-  Character classification macros
+  Character classification macros.
+  The macro CTYPE_DOMAIN is defined in "system.h".
 */
-#define is_digit(c) ((c) >= '0' && (c) <= '9' ? 1 : 0)
-#define is_punct(c) ispunct(c)
-#define is_print(c) (isgraph(c) && ((c) < '0' || (c) > '9'))
+#define is_digit(c) ((unsigned int) (c) - '0' <= 9 ? 1 : 0)
+#define is_punct(c) (CTYPE_DOMAIN((unsigned char)(c)) && ispunct((unsigned char)(c)))
+#define is_print(c) (CTYPE_DOMAIN((unsigned char)(c)) && isgraph((unsigned char)(c)) && ((unsigned int) (c) - '0' > 9))
+#define is_space(c) (CTYPE_DOMAIN((unsigned char)(c)) && isspace((unsigned char)(c)))
 
 /*
   Mathematical functions
 */
 
-int      cmp (const Real *p, const Real *q);
-int      is0 (const Real *u);
+int      cmp (Real p, Real q);
+int      is0 (Real u);
 int      smart_cmp (const Complex* pz1, const Complex* pz2, int flag);
 void     printno (Real u, int m);
 
 extern Real Zero, Inf;
 
-void     init_mpa(void);
-void     mpa_define_epsilon (int iscale, const struct numfmt* pnf);
+void     init_mpa(int iscale);
 
 void     initR (Real* px);
 void     initC (Complex* pz);
-Real     copyR (Real x);
-Complex  copyC (Complex z);
+
+void copyR (Real* dst, Real src);
+void copyC (Complex* dst, Complex src);
 
 #ifdef _MPA_DEBUG
 void     debug_printno (Real u, int m);
-#endif
+#endif /* _MPA_DEBUG */
 
 void     str2R (const char *q, char **endptr, int iscale,
 		const struct numfmt* pnf, Real* pr);
@@ -340,8 +411,17 @@ void     Csub (Complex z1, Complex z2, Complex* pw, int iscale);
 void     delR (Real* px);
 void     delC (Complex* pz);
 
-void     mpa_undef_epsilon (void);
 void     end_mpa(void);
+
+/*
+  Functions used to manipulate lists of threshold specifications (see thrlist.c)
+*/
+
+thrlist thrlist_new (void);
+int thrlist_add (thrlist *plist, const char* def);
+int thrlist_cmp (Real r, thrlist list, unsigned long fieldno1, unsigned long fieldno2);
+void thrlist_dispose (thrlist *plist);
+
 
 /* Shared definitions coming from GNU DIFF
 
@@ -416,6 +496,9 @@ XTERN enum
   IGNORE_ALL_SPACE
 } ignore_white_space;
 
+/* Treat both files as binary files (meaningful only under Doz/Windoz) */
+XTERN bool binary;
+
 /* Nonzero means to not show common lines.  */
 XTERN bool suppress_common_lines;
 
@@ -550,9 +633,13 @@ char* acxnum (const char *str, const struct numfmt* pnf);
 int compare_numeric_strings (const char *str1, const struct numfmt* pnf1,
 			     const char *str2, const struct numfmt* pnf2);
 char* hcxnum (const char *str, const struct numfmt* pnf, hash_value *ph);
+#ifdef USE_GMP
+int mpf_a2num (Real* pr, const char *q, char** endptr, const struct numfmt* pnf);
+#endif /* USE_GMP */
 
 /* side.c */
 void print_sdiff_script (struct change *);
+void print_1overview_line (const char *left, int are_different, const char *right);
 
 /* util.c */
 bool lines_differ (char const *, char const *, int, int, argslist*);
