@@ -1,7 +1,7 @@
 /*
     Numdiff - compare putatively similar files, 
     ignoring small numeric differences
-    Copyright (C) 2005, 2006, 2007, 2008, 2009  Ivano Primi  <ivprimi@libero.it>
+    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  Ivano Primi  <ivprimi@libero.it>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,6 +21,9 @@
 #include<stdio.h>
 #include<stdlib.h> /* for exit() */
 #include<string.h>
+#if defined(HAVE_LOCALECONV) && defined(USE_GMP)
+#include<locale.h>
+#endif /* HAVE_LOCALECONV and USE_GMP */
 
 /*
   The following structure is used to store 
@@ -83,12 +86,17 @@ anum (const char *str, const struct numfmt* pnf)
 {
   char *ptr;
   long digits, strscale;
+  size_t length;
+
   ptr = (char*)str;
   digits = 0;
   strscale = 0;
+  length = strlen(pnf->currency); 
 
   if ( (*ptr == pnf->pos_sign) || (*ptr == pnf->neg_sign) )
     move_ahead(ptr);
+  if ( length > 0 && strncmp (ptr, pnf->currency, length) == 0 )
+    ptr += length; /* Skip the currency name, if specified */
   if (pnf->grouping > 0)
     {
       while ( (is_digit((int)*ptr)) || *ptr == pnf->thsep )
@@ -138,7 +146,7 @@ anum (const char *str, const struct numfmt* pnf)
     move_ahead(ptr), strscale++;
   if (digits+strscale == 0)
     return (char*)str;
-  if (toupper(*ptr) == toupper(pnf->ech) && !isspace (*(ptr+1)))
+  if (TOLOWER(*ptr) == TOLOWER(pnf->ech) && !is_space (*(ptr+1)))
     {
       char *tail;
       long expn;
@@ -181,7 +189,7 @@ acxnum (const char *str, const struct numfmt* pnf)
 	}
       else
 	{
-	  while (isspace (*ptr))
+	  while (is_space (*ptr))
 	    ptr++;
 	  if (*ptr == POS_SIGN || *ptr == NEG_SIGN)
 	    {
@@ -205,8 +213,10 @@ snum (const char *str, const struct numfmt* pnf,
   int nonnull_digit_found = 0;
   long digits, strscale, nzeros, expn;
   char* tail;
+  size_t length;
 
   digits = strscale = 0;
+  length = strlen(pnf->currency); 
   if ( (*ptr == pnf->pos_sign) )
     {
       pcform->sgn = POS_SIGN;
@@ -217,6 +227,8 @@ snum (const char *str, const struct numfmt* pnf,
       pcform->sgn = NEG_SIGN;
       move_ahead(ptr);
     }
+  if ( length > 0 && strncmp (ptr, pnf->currency, length) == 0 )
+    ptr += length; /* Skip the currency name, if specified */
   if (pnf->grouping > 0)
     {
       while ( (is_digit((int)*ptr)) || *ptr == pnf->thsep )
@@ -290,7 +302,7 @@ snum (const char *str, const struct numfmt* pnf,
       pcform->ipart = pcform->dpart = NULL;
       pcform->sgn = 0;
     }
-  if (toupper(*ptr) == toupper(pnf->ech) && !isspace (*(ptr+1)))
+  if (TOLOWER(*ptr) == TOLOWER(pnf->ech) && !is_space (*(ptr+1)))
     {
       expn = strtol (ptr + 1, &tail, 10);
       if (tail != ptr + 1)
@@ -476,7 +488,7 @@ scxnum (const char *str, const struct numfmt* pnf,
 	    complex number.
 	  */
 	  *pre = cform1;
-	  while (isspace (*ptr))
+	  while (is_space (*ptr))
 	    ptr++;
 	  if (*ptr == POS_SIGN || *ptr == NEG_SIGN)
 	    {
@@ -766,3 +778,116 @@ hcxnum (const char *str, const struct numfmt* pnf, hash_value *ph)
   return endptr;
 }
 
+#ifdef USE_GMP
+
+int mpf_a2num (Real* pr, const char *q, char** endptr, const struct numfmt* pnf)
+{
+  struct canform cform;
+  size_t slen = strlen (q);
+
+  cform.sgn   = 0;
+  cform.ipart = stralloc (slen);
+  cform.dpart = stralloc (slen);
+  cform.expn  = 0;  
+  *endptr = snum (q, pnf, &cform);
+  if (*endptr != q)
+    {
+      char *ptr, *str; 
+      long e;
+      size_t len, explen;
+#ifdef HAVE_LOCALECONV
+      struct lconv *plconv = localeconv(); 
+      char* dec_point = plconv->decimal_point;
+#else /* not HAVE_LOCALECONV */ 
+      char* dec_point = ".";
+#endif /* not HAVE_LOCALECONV */ 
+
+      if (cform.ipart == NULL || cform.dpart == NULL)
+	{
+	  /*
+	    The number contained in the string Q is zero
+	  */
+	  mpf_set_ui (*pr, 0);
+	  return 0;
+	}
+      normalize (&cform);
+      e = cform.expn >= 0 ? cform.expn : -cform.expn;
+      for (len = 0; e != 0; len++, e /= 10);
+      explen = len;
+      if (cform.expn < 0)
+	/* We need space for the minus sign in front of the exponent */
+	len++; 
+      if (cform.sgn == NEG_SIGN)
+	/* We need space for the minus sign in front of the whole number */
+	len++;
+      /* We need space also for the decimal point and the exponent letter */
+      len += strlen (cform.ipart) + strlen (cform.dpart) + strlen(dec_point) + 1;
+      ptr = str = stralloc (len);
+
+      if (cform.sgn == NEG_SIGN)
+	{
+	  *ptr = NEG_SIGN;
+	  ptr++;
+	}
+      if (*cform.ipart != '\0')
+	{
+	  *ptr = *cform.ipart;
+	  strcat (str, dec_point);
+	  strcat (str, cform.ipart+1);
+	  strcat (str, cform.dpart);
+	}
+      else
+	{
+	  *ptr = *cform.dpart;
+	  strcat (str, dec_point);
+	  strcat (str, cform.dpart+1);
+	}
+      if (explen > 0)
+	{
+	  for (ptr++; *ptr != '\0'; ptr++);
+	  *ptr = ECH;
+	  if (cform.expn < 0)
+	    {
+	      ptr++;
+	      *ptr = NEG_SIGN;
+	    }
+	  ptr += explen;
+	  for (e = cform.expn >= 0 ? cform.expn : -cform.expn; e != 0; ptr--, e /= 10)
+	    *ptr = CHAR_ZERO + e % 10;
+	}
+      if ( mpf_set_str (*pr, str, 10) == -1 )
+	{
+	  /* This should never happen.  If mpf_set_str() returns -1, then */
+	  /* there is something wrong with the code above.                */
+	  fprintf (stderr,
+		   _("The string \"%s\"\nis not a valid number, the execution of the program ends now\n"),
+		   str);
+	  free ((void*) cform.ipart);
+	  free ((void*) cform.dpart);
+	  free ((void*) str);
+	  exit (EXIT_TROUBLE);
+	}
+      else
+	{
+	  free ((void*) cform.ipart);
+	  free ((void*) cform.dpart);
+	  free ((void*) str);
+	  return 0;
+	}
+    } /* *endptr != q */
+  else
+    {
+      /*
+	We have read no valid number, then
+	free the previously allocated memory,
+	set the value pointed to by PR
+	to zero and return -1
+       */
+      free ((void*) cform.ipart);
+      free ((void*) cform.dpart);
+      mpf_set_ui (*pr, 0);
+      return -1;
+    } /* *endptr == q */
+}
+
+#endif /* USE_GMP */
