@@ -1,7 +1,7 @@
 /*
     Numdiff - compare putatively similar files, 
     ignoring small numeric differences
-    Copyright (C) 2005, 2006, 2007, 2008, 2009  Ivano Primi  <ivprimi@libero.it>
+    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  Ivano Primi  <ivprimi@libero.it>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -38,7 +38,7 @@
 void print_version (const char* progname)
 {
   printf ("%s %s\n", progname, VERSION);
-  printf ("Copyright (C) 2005, 2006, 2007, 2008, 2009  %s <ivprimi@libero.it>\n", 
+  printf ("Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012  %s <ivprimi@libero.it>\n", 
 	  /* TRANSLATORS: This is a proper name.  See the gettext
 	     manual, section Names.
 	     Pronounciation is like "evaa-no pree-me".  */
@@ -48,39 +48,59 @@ License GPLv3+: GNU GPL version 3 or later,\n\
 see <http://gnu.org/licenses/gpl.html>.\n\
 This is free software: you are free to change and redistribute it.\n\
 There is NO WARRANTY, to the extent permitted by law.\n"));
-  printf ("\n%s:\n", _("Built with support for Multiple Precision Arithmetic"));
-  printf ("%s\n%s\n\n", 
-	  _("Epsilon = 1.0e-##, where ## is the precision set by the user"),
-	  _("Decimal digits of accuracy =  precision set by the user"));
+#ifdef USE_GMP
+  printf ("\n%s %s.\n", _("The software has been linked against\n\
+the GNU Multiple Precision Arithmetic Library,\n\
+version number"), gmp_version);
+#else  /* not USE_GMP */
+  printf ("\n%s.\n", _("The software has been built with\n\
+its own internal support for multiple precision arithmetic"));
+#endif /* not USE_GMP */
 }
 
 void print_help (const char* progname)
 {
   puts (_("Usage:"));
   printf ("%s -h|--help|-v|--version   %s\n\n", progname, _("or"));
-  printf ("%s %s\n", progname, "[-s IFS][-a MAXERR][-r MAXERR][-2][-# PREC][-P][-N][-I][-d C1C2][-t C1C2][-g N1N2][-p C1C2][-n C1C2][-e C1C2][-i C1C2][-X 1:INT][-X 2:INT][-E][-D][-b][-V][-q][-S][-z 1:INT][-z 2:INT][-Z 1:INT][-Z 2:INT][-m][-H][-f[NUM]][-T][-l PATH][-o PATH] FILE1 FILE2");
-  printf (_("\nINT stays for a positive integer value or for a range of integer values,\nlike 1-, 3-5 or -7\n"));
+  printf ("%s %s\n", progname, "[-s IFS][-a THRVAL[:RANGE|:RANGE1:RANGE2]][-r THRVAL[:RANGE|:RANGE1:RANGE2]][-2][-F NUM][-# NUM][-P][-N][-I][-c CURRNAME][-d C1C2][-t C1C2][-g N1N2][-p C1C2][-n C1C2][-e C1C2][-i C1C2][-X 1:RANGE][-X 2:RANGE][-E][-D][-b][-V][-O[NUM]][-q][-S][-z 1:RANGE][-z 2:RANGE][-Z 1:RANGE][-Z 2:RANGE][-m][-H][-f[NUM]][-T][-B][-l PATH][-o PATH] FILE1 FILE2");
+  printf (_("\nCompare putatively similar files line by line and field by field,\nignoring small numeric differences or/and different numeric formats.\n\n"));
+  printf (_("RANGE, RANGE1 and RANGE2 stay for a positive integer value or\nfor a range of integer values, like 1-, 3-5 or -7.\n"));
+  printf ("%s\n%s\n%s\n\n%s\n\n",
+	  _("The two arguments after the options are the names of the files to compare."),
+	  _("The complete paths of the files should be given,\na directory name is not accepted."),
+	  _("The given paths cannot refer to the same file\nbut one of them can be \"-\", which refers to stdin."),
+	  _("Exit status: 1 if files differ, 0 if they are equal, -1 (255) in case of error"));
   /* %%% */
-  printf (_("\nCompare putatively similar files line by line and field by field,\nignoring small numeric differences or/and different numeric formats\n\n"));
-  printf ("-s, --separator=IFS\n    %s\n    %s\n",
+  printf ("-s, --separator=IFS\n    %s\n    %s\n    %s\n",
 	  _("Specify the set of characters to use\n    to split the input lines into fields"),
-	  _("(The default set of characters is space, tab and newline)"));
-  printf ("-a, --absolute-tolerance=MAXERR\n    %s\n    %s\n", 
-	  _("Specify the maximum absolute difference permitted\n    before that two numeric fields are regarded as different"),
-	  _("(The default value is zero)"));
-  printf ("-r, --relative-tolerance=MAXERR\n    %s\n    %s\n", 
-	  _("Specify the maximum relative difference permitted\n    before that two numeric fields are regarded as different"),
-	  _("(The default value is zero)"));
+	  _("(The default set of characters is space, tab and newline)."),
+	  _("If IFS is prefixed with 1: or 2: then use the given character set\n    only for the lines from the first or the second file respectively"));
+  printf ("-a, --absolute-tolerance=THRVAL[:RANGE|:RANGE1:RANGE2]\n    %s\n    %s\n    %s\n", 
+	  _("Set to THRVAL the maximum absolute difference permitted\n    before that two numeric fields are regarded as different\n    (The default value is zero)."),
+	  _("If a RANGE is given, use the specified\n    threshold only when comparing fields whose positions lie in RANGE."),
+	  _("If both RANGE1 and RANGE2 are given and have the same length,\n    then use the specified threshold when comparing a field of FILE1\n    lying in RANGE1 with the corresponding field of FILE2 in RANGE2"));
+  printf ("-r, --relative-tolerance=THRVAL[:RANGE|:RANGE1:RANGE2]\n    %s\n    %s\n    %s\n", 
+	  _("Set to THRVAL the maximum relative difference permitted\n    before that two numeric fields are regarded as different\n    (The default value is zero)."),
+	  _("If a RANGE is given, use the specified\n    threshold only when comparing fields whose positions lie in RANGE."),
+	  _("If both RANGE1 and RANGE2 are given and have the same length,\n    then use the specified threshold when comparing a field of FILE1\n    lying in RANGE1 with the corresponding field of FILE2 in RANGE2"));
   printf ("-2, --strict\n    %s\n",
-	  _("Order that two numerical values are regarded as equal only if\n    both absolute and relative difference do not exceed\n    the corresponding tolerance threshold"));
-  printf ("-#, --digits=PREC\n    %s\n",
-	  _("Specify the number of digits in the significands\n    used in multiple-precision arithmetic"));
+	  _("Consider two numerical values as equal only if\n    both absolute and relative difference do not exceed\n    the corresponding tolerance threshold"));
+  printf ("-F, --formula=NUM\n    %s\n    %s\n    %s\n    %s\n",
+	  _("Use the formula indicated by NUM to compute the relative errors."),
+	  _("If \'NUM\' is 0 use the classic formula."),
+	  _("If \'NUM\' is 1 compute the relative errors by considering\n    the values in FILE1 as sample values."),
+	  _("If \'NUM\' is 2 compute the relative errors by considering\n    the values in FILE2 as sample values."));
+  printf ("-#, --digits=NUM\n    %s\n",
+	  _("Set to NUM the number of digits in the significands\n    used in multiple precision arithmetic"));
   printf ("-P, --positive-differences\n    %s\n",
 	  _("Ignore all differences due to numeric fields of the second file that\n    are less than the corresponding numeric fields in the first file"));
   printf ("-N, --negative-differences\n    %s\n",
 	  _("Ignore all differences due to numeric fields of the second file that\n    are greater than the corresponding numeric fields in the first file"));
   printf ("-I, --ignore-case\n    %s\n",
 	  _("Ignore changes in case while doing literal comparisons"));
+  printf ("-c, --currency=CURRNAME\n    %s\n    %s\n",
+	  _("Set to CURRNAME the currency name for the two files to compare."),
+	  _("CURRNAME must be prefixed with 1: or 2: to specify the\n    currency name only for the first or the second file"));
   printf ("-d, --decimal-point=C1C2\n    %s\n",
 	  _("Specify the characters representing the decimal point\n    in the two files to compare"));
   printf ("-t, --thousands-separator=C1C2\n    %s\n",
@@ -95,9 +115,9 @@ void print_help (const char* progname)
 	  _("Specify the exponent letters\n    used in the two files to compare"));
   printf ("-i, --imaginary-unit=C1C2\n    %s\n",
 	  _("Specify the characters representing the imaginary unit\n    in the two files to compare"));
-  printf ("-X, --exclude=1:INT\n    %s\n",
+  printf ("-X, --exclude=1:RANGE\n    %s\n",
 	  _("Select the fields of the first file that have to be ignored"));
-  printf ("-X, --exclude=2:INT\n    %s\n",
+  printf ("-X, --exclude=2:RANGE\n    %s\n",
 	  _("Select the fields of the second file that have to be ignored"));
   printf ("-E, --essential\n    %s\n",
 	  _("While printing the differences between the two compared files\n    show only the numerical ones"));
@@ -107,17 +127,22 @@ void print_help (const char* progname)
 	  _("Suppress all messages concerning the differences discovered\n    in the structures of the two files"));
   printf ("-V, --verbose\n    %s\n",
 	  _("For every couple of lines which differ in at least one field print\n    an header to show how these lines appear in the two compared files"));
+  printf ("-O, --overview[=NUM]\n    %s\n    %s\n    %s\n    %s\n",
+	  _("Display a side by side difference listing of the two files\n    showing which lines are present only in one file, which\n    lines are present in both files but with one or more differing fields,\n    and which lines are identical."),
+	  _("If \'NUM\' is zero or is not specified, output at most 130 columns per line."),
+	  _("If \'NUM\' is a positive number, output at most \'NUM\' columns per line."),
+	  _("If \'NUM\' is a negative number, do not output common lines\n    and display at most -\'NUM\' columns per line."));
   printf ("-q, --quiet, --silent\n    %s\n",
 	  _("Suppress all the standard output"));
   printf ("-S, --statistics\n    %s\n",
 	  _("Add some statistics to the standard output"));
-  printf ("-z, --blur-if-numerical=1:INT\n    %s\n",
+  printf ("-z, --blur-if-numerical=1:RANGE\n    %s\n",
 	  _("Select the fields of the first file that have to be\n    blurred during the synchronization procedure\n    only if they turn out to be numeric"));
-  printf ("-z, --blur-if-numerical=2:INT\n    %s\n",
+  printf ("-z, --blur-if-numerical=2:RANGE\n    %s\n",
 	  _("Select the fields of the second file that have to be\n    blurred during the synchronization procedure\n    only if they turn out to be numeric"));
-  printf ("-Z, --blur-unconditionally=1:INT\n    %s\n",
+  printf ("-Z, --blur-unconditionally=1:RANGE\n    %s\n",
 	  _("Select the fields of the first file that have to be\n    unconditionally blurred during the synchronization procedure"));
-  printf ("-Z, --blur-unconditionally=2:INT\n    %s\n",
+  printf ("-Z, --blur-unconditionally=2:RANGE\n    %s\n",
 	  _("Select the fields of the second file that have to be\n    unconditionally blurred during the synchronization procedure"));
   printf ("-m, --minimal\n    %s\n",
 	  _("During synchronization try hard to find a smaller set of changes"));
@@ -126,23 +151,21 @@ void print_help (const char* progname)
   printf ("-f, --test-filter[=NUM]\n    %s\n    %s\n    %s\n    %s\n",
 	  _("Run only the filter and then show the results of its\n    attempt to synchronize the two files."),
 	  _("If \'NUM\' is zero or is not specified, output at most 130 columns per line."),
-	  _("If \'NUM\' is a positive number, output at most\n    \'NUM\' columns per line."),
+	  _("If \'NUM\' is a positive number, output at most \'NUM\' columns per line."),
 	  _("If \'NUM\' is a negative number, do not output common lines\n    and display at most -\'NUM\' columns per line."));
   printf ("-T, --expand-tabs\n    %s\n",
-	  _("Expand tabs to spaces in output while displaying the results of the\n    synchronization procedure (meaningful only together with option -f)"));
+	  _("Expand tabs to spaces in output while displaying the results of the\n    synchronization procedure (meaningful only together with option -O or -f)"));
+  printf ("-B, --binary\n    %s\n",
+	  _("Treat both files as binary files (only meaningful under Doz/Windoz)"));
   printf ("-l, --warnings-to=PATH\n    %s\n",
 	  _("Redirect warning and error messages from stderr to the indicated file"));
   printf ("-o, --output=PATH\n    %s\n",
 	  _("Redirect output from stdout to the indicated file"));
   printf ("-h, --help\n    %s\n", _("Show help message and predefined settings"));
   printf ("-v, --version\n    %s\n", _("Show version number, Copyright, Distribution Terms and NO-Warranty"));
-  printf ("\n%s\n%s\n%s\n%s\n",
-	  _("The two arguments after the options are the names of the files to compare."),
-	  _("The complete paths of the files should be given,\na directory name is not accepted."),
-	  _("They cannot refer to the same file but one of them can be \"-\",\nwhich refers to stdin."),
-	  _("Exit status: 1 if files differ, 0 if they are equal, -1 (255) in case of error"));
   /* %%% */
   puts (_("\n  Default numeric format (for both files to compare):\n"));
+  printf (_("Currency name = \"%s\"\n"), CURRENCY);
   printf (_("Decimal point = `%c\'\n"), DP);
 
   printf (_("Thousands separator = `%c\'\n"), THSEP);
@@ -321,18 +344,34 @@ int fselect (const char* str, unsigned char* mask, int mask_size)
     return -1;
   else
     {
+      /* Remark: internally the field numbers
+	 start from zero, not from one */
       for (n = beg - 1; n <= end - 1; n++)
 	mask[n >> 3] |= 0x80 >> (n & 0x7);
       return 1;
     }
 }
 
+#define VALID_NUMFMT      0
+#define INVALID_NUMFMT   -1
+#define INVALID_CURRENCY -2
+
 static
-int valid_numfmt (const struct numfmt* pnf)
+int is_numfmt_valid (const struct numfmt* pnf)
 {
   char store[NUMFMT_CHARS];
+  char *ptr;
   int i, j;
 
+  for (ptr = pnf->currency; 
+       *ptr != '\0' && 
+	 (!is_digit(*ptr)) &&
+	 *ptr != pnf->dp &&
+	 *ptr != pnf->thsep &&
+	 *ptr != pnf->pos_sign &&
+	 *ptr != pnf->neg_sign; ptr++);
+  if (*ptr != '\0')
+    return INVALID_CURRENCY;
   store[0] = pnf->dp;
   store[1] = pnf->thsep;
   store[2] = pnf->pos_sign;
@@ -343,21 +382,25 @@ int valid_numfmt (const struct numfmt* pnf)
     {
       for (j = i+1; j < NUMFMT_CHARS; j++)
 	if (store[i] == store[j])
-	  return 0;
+	  return INVALID_NUMFMT;
     }
-  return 1;
+  return VALID_NUMFMT;
 }
 
 extern int optind;
 
 int setargs (int argc, char* argv[], argslist *list)
 {
-  const char *optstring = "h2bVqDESIPNz:Z:mHT#:s:a:r:d:t:g:p:n:e:i:f::X:l:o:v";
+  const int mask_size = FIELDMASK_SIZE*8;
+  const char *optstring = "h2F:bBVO::qDESIPNz:Z:mHT#:s:a:r:c:d:t:g:p:n:e:i:f::X:l:o:v";
   struct option long_options[] = {
     {"help",                 0, NULL, 'h'},
     {"strict",               0, NULL, '2'},
+    {"formula",              1, NULL, 'F'},
     {"brief",                0, NULL, 'b'},
+    {"binary",               0, NULL, 'B'},
     {"verbose",              0, NULL, 'V'},
+    {"overview",             2, NULL, 'O'},
     {"quiet",                0, NULL, 'q'},
     {"silent",               0, NULL, 'q'},
     {"dummy",                0, NULL, 'D'},
@@ -375,6 +418,7 @@ int setargs (int argc, char* argv[], argslist *list)
     {"separator",            1, NULL, 's'},
     {"absolute-tolerance",   1, NULL, 'a'},
     {"relative-tolerance",   1, NULL, 'r'},
+    {"currency",             1, NULL, 'c'},
     {"decimal-point",        1, NULL, 'd'},
     {"thousands-separator",  1, NULL, 't'},
     {"group-length",         1, NULL, 'g'},
@@ -391,26 +435,25 @@ int setargs (int argc, char* argv[], argslist *list)
   };
   int option_index=0;
   char *tail;
-  int i, optch, off; 
-  unsigned int t;
+  int i, optch, off, rv; 
+  unsigned int t, file_id;
   long w;
   unsigned char *bitmask;  
-  struct numfmt defaults;
 
   /*
     We start by loading the default values
     for the user settable options.
 
-    The initialization of these variables:
+    The initialization of the variables
 
     list->maxrelerr, list->maxabserr,
     list->Labserr, list->Crelerr, list->Lrelerr, list->Cabserr,
     list->N1abserr, list->N1disperr, list->N2abserr, list->N2disperr 
 
-    to Zero is done within main()
-    through init_mpa_support() .
+    is done within main() through init_mpa_support().
   */
 
+  binary = 0;
   suppress_common_lines = 0;
   ignore_white_space = IGNORE_NO_WHITE_SPACE;
   expand_tabs = 0;
@@ -435,21 +478,9 @@ int setargs (int argc, char* argv[], argslist *list)
   list->nf1.ech = ECH;
   list->nf1.iu = IU;
   list->file1 = list->file2 = NULL;
-  defaults = list->nf2 = list->nf1;
-
-  /*
-    defaults.dp == DP
-    defaults.thsep == THSEP;
-    defaults.grouping == GROUPING;
-    defaults.pos_sign == POS_SIGN;
-    defaults.neg_sign == NEG_SIGN;
-    defaults.ech == ECH;
-    defaults.iu == IU;
-
-    Since 'defaults' is not modified in the following
-    we are sure that it will always contain
-    these default values.
-  */
+  list->nf2 = list->nf1;
+  list->nf1.currency = xstrdup (CURRENCY);
+  list->nf2.currency = xstrdup (CURRENCY);
 
   while ( (optch = getopt_long (argc, argv, optstring, long_options, &option_index)) != -1 )
     {
@@ -461,9 +492,26 @@ int setargs (int argc, char* argv[], argslist *list)
 	case '2':
 	  list->optmask |= _2_MASK;
 	  break;	  
+	case 'F':
+	  if (strncmp ("0", optarg, 1) == 0)
+	    list->relerr_formula = CLASSIC_FORMULA;	      
+	  else if (strncmp ("1", optarg, 1) == 0)
+	    list->relerr_formula = WR_TO_FIRST_FILE;	      
+	  else if (strncmp ("2", optarg, 1) == 0)
+	    list->relerr_formula = WR_TO_SECOND_FILE;
+	  else
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+		       PACKAGE, optch);
+	      return -1;
+	    }
+	  break;
 	case 'b':
 	  list->optmask |= _B_MASK;
 	  break;
+	case 'B':
+	  binary = 1;
+	  break;
 	case 'V':
 	  list->optmask |= _SV_MASK;
 	  break;
@@ -497,7 +545,7 @@ int setargs (int argc, char* argv[], argslist *list)
 		None of the prefixes 1: and 2: has been used,
 		then we have to select fields for both files
 	      */
-	      if (fselect (optarg, list->pblurmask1, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg, list->pblurmask1, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -505,14 +553,14 @@ int setargs (int argc, char* argv[], argslist *list)
 		}
 	      else
 		{
-		  fselect (optarg, list->pblurmask2, FIELDMASK_SIZE*8);
+		  fselect (optarg, list->pblurmask2, mask_size);
 		  list->optmask |= _Z_MASK;
 		}
 	    }
 	  else 
 	    {
 	      bitmask = i == 0 ? list->pblurmask1 : list->pblurmask2;
-	      if (fselect (optarg+2, bitmask, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg+2, bitmask, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -529,7 +577,7 @@ int setargs (int argc, char* argv[], argslist *list)
 		None of the prefixes 1: and 2: has been used,
 		then we have to select fields for both files
 	      */
-	      if (fselect (optarg, list->tblurmask1, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg, list->tblurmask1, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -537,14 +585,14 @@ int setargs (int argc, char* argv[], argslist *list)
 		}
 	      else
 		{
-		  fselect (optarg, list->tblurmask2, FIELDMASK_SIZE*8);
+		  fselect (optarg, list->tblurmask2, mask_size);
 		  list->optmask |= _SZ_MASK;
 		}
 	    }
 	  else 
 	    {
 	      bitmask = i == 0 ? list->tblurmask1 : list->tblurmask2;
-	      if (fselect (optarg+2, bitmask, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg+2, bitmask, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -621,42 +669,78 @@ int setargs (int argc, char* argv[], argslist *list)
 	    list->optmask |= _S_MASK;
 	  break;
 	case 'a':
-	  delR (&list->maxabserr); /* To avoid unpleasant memory leaks */
-	  str2R (optarg, &tail, ISCALE, &defaults, &list->maxabserr);
-
-	  if (*tail != '\0')
+	  rv = thrlist_add (&list->maxabserr, optarg);
+	  if (rv == THRLIST_INVALID_FORMAT)
 	    {
-	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
 		       PACKAGE, optch);
+	      fprintf (stderr, _("  The format specification has not been respected\n"));
+	      return -1;
+	    }
+	  else if (rv == THRLIST_INVALID_RANGES)
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
+		       PACKAGE, optch);
+	      fprintf (stderr, _("  The specified ranges do not have the same length\n"));
 	      return -1;
 	    }
 	  else
 	    list->optmask |= _A_MASK;
 	  break;
 	case 'r':
-	  delR (&list->maxrelerr); /* To avoid unpleasant memory leaks */
-	  if (*optarg == '1' && *(optarg+1) ==':')
+	  rv = thrlist_add (&list->maxrelerr, optarg);
+	  if (rv == THRLIST_INVALID_FORMAT)
 	    {
-	      list->relerr_formula = WR_TO_FIRST_FILE;
-	      str2R (optarg+2, &tail, ISCALE, &defaults, &list->maxrelerr);
-	    }
-	  else if (*optarg == '2' && *(optarg+1) ==':')
-	    {
-	      list->relerr_formula = WR_TO_SECOND_FILE;
-	      str2R (optarg+2, &tail, ISCALE, &defaults, &list->maxrelerr);
-	    }
-	  else
-	    str2R (optarg, &tail, ISCALE, &defaults, &list->maxrelerr);
-
-	  if (*tail != '\0')
-	    {
-	      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
 		       PACKAGE, optch);
+	      fprintf (stderr, _("  The format specification has not been respected\n"));
+	      return -1;
+	    }
+	  else if (rv == THRLIST_INVALID_RANGES)
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
+		       PACKAGE, optch);
+	      fprintf (stderr, _("  The specified ranges do not have the same length\n"));
 	      return -1;
 	    }
 	  else
 	    list->optmask |= _R_MASK;
 	  break;
+	case 'c':
+	  file_id = 0x0;
+	  if (*optarg == '1' && *(optarg+1) == ':')
+	    {
+	      if ((list->nf1.currency))
+		free((void*)list->nf1.currency);
+	      list->nf1.currency = xstrdup (optarg+2);
+              file_id = 0x1;
+	    }
+	  else if (*optarg == '2' && *(optarg+1) == ':')
+	    {
+	      if ((list->nf2.currency))
+		free((void*)list->nf2.currency);
+	      list->nf2.currency = xstrdup (optarg+2);
+	      file_id = 0x2;
+	    }
+	  else
+	    {
+	      if ((list->nf1.currency))
+		free((void*)list->nf1.currency);
+	      list->nf1.currency = xstrdup (optarg);
+	      if ((list->nf2.currency))
+		free((void*)list->nf2.currency);
+	      list->nf2.currency = xstrdup (optarg);
+	      file_id = 0x3;
+	    }
+	  if ( (strlen(list->nf1.currency) == 0 && ((file_id & 0x3) == 0x1))
+	    || (strlen(list->nf2.currency) == 0 && ((file_id & 0x3) == 0x2)) )
+	    {
+	      fprintf (stderr, _("%s: invalid argument after `-%c\' option:\n"),
+		       PACKAGE, optch);
+	      fprintf (stderr, _("  you have missed to specify the currency name\n"));
+	      return -1;
+	    }
+	  break;
 	case 'd':
 	case 't':
 	case 'g':
@@ -672,48 +756,60 @@ int setargs (int argc, char* argv[], argslist *list)
 	    }
 	  break;
 	case 'f':
-	  if(!optarg)
+	case 'O':
+	  if (!(list->optmask & (_F_MASK | _SO_MASK)))
 	    {
-	      /* There is no optional argument, then set */
-	      /* 'w' to 'DEF_ATMOST_NCOLS'.              */
-	      list->optmask |= _F_MASK;
-	      w = DEF_ATMOST_NCOLS;
-	    }
-	  else
-	    {
-	      /* An argument follows */
-	      w = strtol (optarg, &tail, 10);
-	      /* If the argument of the option is not a valid number, */
-	      /* then exit after printing a suitable error message.   */
-	      if (*tail != '\0')
+	      if(!optarg)
 		{
-		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
-			   PACKAGE, optch);
-		  return -1;
+		  /* There is no optional argument, then set */
+		  /* 'w' to 'DEF_ATMOST_NCOLS'.              */
+		  if (optch == 'f')
+		    list->optmask |= _F_MASK;
+		  else
+		    list->optmask |= _SO_MASK;
+		  w = DEF_ATMOST_NCOLS;
 		}
 	      else
-		list->optmask |= _F_MASK;
-	      /* Otherwise you have to set 'w' appropriately. */
-	      /* If the given argument is less than -MAX_ATMOST_NCOLS */
-	      /* then set 'w' to 'DEF_ATMOST_NCOLS' and 'suppress_common_lines' to 'TRUE'. */
-	      if (w < -MAX_ATMOST_NCOLS)
 		{
-		  w = DEF_ATMOST_NCOLS;
-		  suppress_common_lines = 1;
-		}
-	      /* If the argument were negative, then remove the sign */
-	      /* and set 'suppress_common_lines' to 'TRUE'.          */
-	      if (w < 0)
-		{
-		  w *= -1;
-		  suppress_common_lines = 1;
-		}
-	      /* If the given argument is too small or too big in absolute value, */
-	      /* then set 'w' to 'DEF_ATMOST_NCOLS'.                              */
-	      if (w < MIN_ATMOST_NCOLS || w > MAX_ATMOST_NCOLS)
-		w = DEF_ATMOST_NCOLS;
-	      /* Otherwise leave 'w' set to the value of the argument. */
-	    } /* end optarg != 0 */
+		  /* An argument follows */
+		  w = strtol (optarg, &tail, 10);
+		  /* If the argument of the option is not a valid number, */
+		  /* then exit after printing a suitable error message.   */
+		  if (*tail != '\0')
+		    {
+		      fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
+			       PACKAGE, optch);
+		      return -1;
+		    }
+		  else
+		    {
+		      if (optch == 'f')
+			list->optmask |= _F_MASK;
+		      else
+			list->optmask |= _SO_MASK;
+		    }
+		  /* Otherwise you have to set 'w' appropriately. */
+		  /* If the given argument is less than -MAX_ATMOST_NCOLS */
+		  /* then set 'w' to 'DEF_ATMOST_NCOLS' and 'suppress_common_lines' to 'TRUE'. */
+		  if (w < -MAX_ATMOST_NCOLS)
+		    {
+		      w = DEF_ATMOST_NCOLS;
+		      suppress_common_lines = 1;
+		    }
+		  /* If the argument were negative, then remove the sign */
+		  /* and set 'suppress_common_lines' to 'TRUE'.          */
+		  if (w < 0)
+		    {
+		      w *= -1;
+		      suppress_common_lines = 1;
+		    }
+		  /* If the given argument is too small or too big in absolute value, */
+		  /* then set 'w' to 'DEF_ATMOST_NCOLS'.                              */
+		  if (w < MIN_ATMOST_NCOLS || w > MAX_ATMOST_NCOLS)
+		    w = DEF_ATMOST_NCOLS;
+		  /* Otherwise leave 'w' set to the value of the argument. */
+		} /* end optarg != 0 */
+	    }
 	  break;
 	case 'X':
 	  if ( (i = strncmp (optarg, "1:", 2)) && (strncmp (optarg, "2:", 2)) )
@@ -722,7 +818,7 @@ int setargs (int argc, char* argv[], argslist *list)
 		None of the prefixes 1: and 2: has been used,
 		then we have to select fields for both files
 	      */
-	      if (fselect (optarg, list->ghostmask1, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg, list->ghostmask1, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -730,14 +826,14 @@ int setargs (int argc, char* argv[], argslist *list)
 		}
 	      else
 		{
-		  fselect (optarg, list->ghostmask2, FIELDMASK_SIZE*8);
+		  fselect (optarg, list->ghostmask2, mask_size);
 		  list->optmask |= _SX_MASK;
 		}
 	    }
 	  else 
 	    {
 	      bitmask = i == 0 ? list->ghostmask1 : list->ghostmask2;
-	      if (fselect (optarg+2, bitmask, FIELDMASK_SIZE*8) <= 0)
+	      if (fselect (optarg+2, bitmask, mask_size) <= 0)
 		{
 		  fprintf (stderr, _("%s: invalid argument after `-%c\' option\n"),
 			   PACKAGE, optch);
@@ -788,6 +884,8 @@ int setargs (int argc, char* argv[], argslist *list)
     list->output_mode = OUTMODE_BRIEF;
   if (list->optmask & _B_MASK && list->optmask & _SV_MASK)
     list->output_mode = OUTMODE_COINCISE;
+  if ( list->optmask & _SO_MASK )
+    list->output_mode = OUTMODE_OVERVIEW;
   if ( list->optmask & _Q_MASK )
     list->output_mode = OUTMODE_QUIET;
 
@@ -796,7 +894,7 @@ int setargs (int argc, char* argv[], argslist *list)
       print_help (PACKAGE);
       return -1;
     }
-  else if ( !valid_numfmt(&list->nf1) )
+  else if ( (rv = is_numfmt_valid(&list->nf1)) == INVALID_NUMFMT )
     {
       fprintf (stderr, 
 	       _("The numeric format specified for the first file is illegal,\n"));
@@ -813,7 +911,21 @@ int setargs (int argc, char* argv[], argslist *list)
 	       list->nf1.iu);
       return -1;
     }
-  else if ( !valid_numfmt(&list->nf2) )
+  else if ( rv == INVALID_CURRENCY )
+    {
+      fprintf (stderr, 
+	       _("The numeric format specified for the first file is illegal:\n"));
+      fprintf (stderr,
+	       _("the name of the currency may not contain digits,\n"));
+      fprintf (stderr,
+	       _("the symbol for the leading positive sign (`%c\'),\n"), list->nf1.pos_sign);
+      fprintf (stderr,
+	       _("the symbol for the leading negative sign (`%c\'),\n"), list->nf1.neg_sign);
+      fprintf (stderr,
+	       _("the decimal point (`%c\'), or the thousands separator (`%c\')\n"), list->nf1.dp, list->nf1.thsep);
+      return -1;
+    }
+  else if ( (rv = is_numfmt_valid(&list->nf2)) == INVALID_NUMFMT )
     {
       fprintf (stderr, 
 	       _("The numeric format specified for the second file is illegal,\n"));
@@ -830,6 +942,20 @@ int setargs (int argc, char* argv[], argslist *list)
 	       list->nf2.iu);
       return -1;
     }
+  else if ( rv == INVALID_CURRENCY )
+    {
+      fprintf (stderr, 
+	       _("The numeric format specified for the second file is illegal:\n"));
+      fprintf (stderr,
+	       _("the name of the currency may not contain digits,\n"));
+      fprintf (stderr,
+	       _("the symbol for the leading positive sign (`%c\'),\n"), list->nf2.pos_sign);
+      fprintf (stderr,
+	       _("the symbol for the leading negative sign (`%c\'),\n"), list->nf2.neg_sign);
+      fprintf (stderr,
+	       _("the decimal point (`%c\'), or the thousands separator (`%c\')\n"), list->nf2.dp, list->nf2.thsep);
+      return -1;
+    }
   else
     {
       if( !(list->optmask & (_H_MASK | _V_MASK)) )
